# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase
from mock import MagicMock

from qrl.core import config
from qrl.core.TransactionMetadata import TransactionMetadata
from qrl.core.misc import logger, db
from qrl.core.State import State
from qrl.core.txs.TransferTransaction import TransferTransaction
from qrl.core.Block import Block

from tests.misc.helper import set_qrl_dir, get_alice_xmss, get_some_address

logger.initialize_default()


class TestTransactionMetadata(TestCase):
    def setUp(self):
        with set_qrl_dir('no_data'):
            self.state = State()
        self.m_db = MagicMock(name='mock DB', autospec=db.DB)

    def test_rollback_tx_metadata(self):
        alice_xmss = get_alice_xmss()

        tx1 = TransferTransaction.create(addrs_to=[get_some_address(1), get_some_address(2)],
                                         amounts=[1, 2],
                                         message_data=None,
                                         fee=0,
                                         xmss_pk=alice_xmss.pk)

        block = Block.create(dev_config=config.dev,
                             block_number=5,
                             prev_headerhash=b'',
                             prev_timestamp=10,
                             transactions=[tx1],
                             miner_address=b'',
                             seed_height=0,
                             seed_hash=None)

        TransactionMetadata.update_tx_metadata(self.state, block=block, batch=None)

        tx_metadata = TransactionMetadata.get_tx_metadata(self.state, tx1.txhash)

        self.assertEqual(tx_metadata[0].to_json(), tx1.to_json())
        TransactionMetadata.rollback_tx_metadata(self.state, block, None)
        self.assertIsNone(TransactionMetadata.get_tx_metadata(self.state, tx1.txhash))

    def test_update_tx_metadata(self):
        alice_xmss = get_alice_xmss()
        tx = TransferTransaction.create(addrs_to=[get_some_address(1), get_some_address(2)],
                                        amounts=[1, 2],
                                        message_data=None,
                                        fee=0,
                                        xmss_pk=alice_xmss.pk)
        block_number = 5
        TransactionMetadata.put_tx_metadata(self.state, tx, block_number, 10000, None)

        tx_metadata = TransactionMetadata.get_tx_metadata(self.state, tx.txhash)
        self.assertEqual(tx_metadata[0].to_json(), tx.to_json())
        self.assertEqual(tx_metadata[1], block_number)

    def test_remove_tx_metadata(self):
        self.assertIsNone(TransactionMetadata.get_tx_metadata(self.state, b'test1'))

        alice_xmss = get_alice_xmss()
        tx = TransferTransaction.create(addrs_to=[get_some_address(1), get_some_address(2)],
                                        amounts=[1, 2],
                                        message_data=None,
                                        fee=0,
                                        xmss_pk=alice_xmss.pk)
        block_number = 5
        TransactionMetadata.put_tx_metadata(self.state, tx, block_number, 10000, None)

        tx_metadata = TransactionMetadata.get_tx_metadata(self.state, tx.txhash)
        self.assertEqual(tx_metadata[0].to_json(), tx.to_json())
        self.assertEqual(tx_metadata[1], block_number)

        TransactionMetadata.remove_tx_metadata(self.state, tx, None)
        self.assertIsNone(TransactionMetadata.get_tx_metadata(self.state, tx.txhash))

    def test_put_tx_metadata(self):
        self.assertIsNone(TransactionMetadata.get_tx_metadata(self.state, b'test1'))

        alice_xmss = get_alice_xmss()
        tx = TransferTransaction.create(addrs_to=[get_some_address(1), get_some_address(2)],
                                        amounts=[1, 2],
                                        message_data=None,
                                        fee=0,
                                        xmss_pk=alice_xmss.pk)
        block_number = 5
        TransactionMetadata.put_tx_metadata(self.state, tx, block_number, 10000, None)

        tx_metadata = TransactionMetadata.get_tx_metadata(self.state, tx.txhash)
        self.assertEqual(tx_metadata[0].to_json(), tx.to_json())
        self.assertEqual(tx_metadata[1], block_number)

    def test_get_tx_metadata(self):
        self.assertIsNone(TransactionMetadata.get_tx_metadata(self.state, b'test1'))

        alice_xmss = get_alice_xmss()
        tx = TransferTransaction.create(addrs_to=[get_some_address(1), get_some_address(2)],
                                        amounts=[1, 2],
                                        message_data=None,
                                        fee=0,
                                        xmss_pk=alice_xmss.pk)
        block_number = 5
        timestamp = 10000
        TransactionMetadata.put_tx_metadata(self.state, tx, block_number, timestamp, None)

        tx_metadata = TransactionMetadata.get_tx_metadata(self.state, tx.txhash)
        self.assertEqual(tx_metadata[0].to_json(), tx.to_json())
        self.assertEqual(tx_metadata[1], block_number)



from unittest import TestCase

from mock import Mock

from qrl.core import config
from qrl.core.Indexer import Indexer
from qrl.core.misc import logger
from qrl.core.State import State
from qrl.core.StateContainer import StateContainer
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.ChainManager import ChainManager
from qrl.core.txs.TokenTransaction import TokenTransaction
from qrl.generated import qrl_pb2
from tests.misc.helper import get_alice_xmss, get_bob_xmss, get_slave_xmss, set_qrl_dir

logger.initialize_default()


class TestTokenTransactionStateChanges(TestCase):
    def setUp(self):
        with set_qrl_dir('no_data'):
            self.state = State()

        self.alice = get_alice_xmss()
        self.bob = get_bob_xmss()

        alice_address_state = OptimizedAddressState.get_default(self.alice.address)
        alice_address_state.pbdata.balance = 100
        self.addresses_state = {
            self.alice.address: alice_address_state,
            self.bob.address: OptimizedAddressState.get_default(self.bob.address)
        }

        self.params = {
            "symbol": b'QRL',
            "name": b'Quantum Resistant Ledger',
            "owner": self.alice.address,
            "decimals": 15,
            "initial_balances": [],
            "fee": 1,
            "xmss_pk": self.alice.pk
        }

        self.unused_chain_manager_mock = Mock(autospec=ChainManager, name='unused ChainManager')

    def test_apply_token_txn(self):
        """
        Alice creates a token. Obviously, she gives herself some of this token.
        But she also gives Bob some tokens too.
        """
        initial_balances = [qrl_pb2.AddressAmount(address=self.alice.address, amount=1000),
                            qrl_pb2.AddressAmount(address=self.bob.address, amount=2000)]
        self.params["initial_balances"] = initial_balances

        tx = TokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = dict(self.addresses_state)
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)

        # According to the State, Alice has 100 coins, and Bob has 0 coins.
        # After applying the Transaction, Alice and Bob should have 1000 tokens, and Alice's balance should be 99.
        # AddressState.transaction_hashes now also reference the TokenTransaction that created the Tokens.
        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))
        tx.apply(self.state, state_container)
        self.assertTrue(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))
        self.assertEqual(addresses_state[self.alice.address].balance, 99)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)

        self.assertEqual(2, len(state_container.paginated_tx_hash.key_value))

        storage_key = state_container.paginated_tx_hash.generate_key(self.alice.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([tx.txhash], state_container.paginated_tx_hash.key_value[storage_key])

        storage_key = state_container.paginated_tx_hash.generate_key(self.bob.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([tx.txhash], state_container.paginated_tx_hash.key_value[storage_key])

        self.assertEqual(2, len(state_container.tokens.data))
        self.assertIn((self.alice.address, tx.txhash), state_container.tokens.data)
        self.assertEqual(1000, state_container.tokens.data[(self.alice.address, tx.txhash)].balance)
        self.assertIn((self.bob.address, tx.txhash), state_container.tokens.data)
        self.assertEqual(2000, state_container.tokens.data[(self.bob.address, tx.txhash)].balance)

    def test_apply_state_changes_owner_not_in_address_state(self):
        """
        In this case, Alice didn't give herself any tokens. How generous! She gave them all to Bob.
        """
        initial_balances = [qrl_pb2.AddressAmount(address=self.bob.address, amount=1000)]
        self.params["initial_balances"] = initial_balances

        tx = TokenTransaction.create(**self.params)
        tx.sign(self.alice)

        # Signing the TX also generates the txhash, which we need to generate the AddressState properly.
        addresses_state = dict(self.addresses_state)
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        tx.apply(self.state, state_container)

        self.assertEqual(addresses_state[self.alice.address].balance, 99)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)

        self.assertEqual(2, len(state_container.paginated_tx_hash.key_value))

        storage_key = state_container.paginated_tx_hash.generate_key(self.alice.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([tx.txhash], state_container.paginated_tx_hash.key_value[storage_key])

        storage_key = state_container.paginated_tx_hash.generate_key(self.bob.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([tx.txhash], state_container.paginated_tx_hash.key_value[storage_key])

        self.assertEqual(1, len(state_container.tokens.data))
        self.assertIn((self.bob.address, tx.txhash), state_container.tokens.data)
        self.assertEqual(1000, state_container.tokens.data[(self.bob.address, tx.txhash)].balance)

    def test_apply_token_txn_signed_by_slave_xmss(self):
        """
        Alice creates a token, gives herself and Bob some tokens.
        But she uses a XMSS slave to sign it.
        """
        initial_balances = [qrl_pb2.AddressAmount(address=self.alice.address, amount=1000),
                            qrl_pb2.AddressAmount(address=self.bob.address, amount=2000)]
        slave = get_slave_xmss()
        self.params["initial_balances"] = initial_balances
        self.params["xmss_pk"] = slave.pk
        self.params["master_addr"] = self.alice.address
        tx = TokenTransaction.create(**self.params)
        tx.sign(slave)

        # Now that we have the Slave XMSS address, we should add it to AddressState so that apply_state_changes()
        # can do something with it
        addresses_state = dict(self.addresses_state)
        addresses_state[slave.address] = OptimizedAddressState.get_default(slave.address)
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        tx.apply(self.state, state_container)

        self.assertEqual(addresses_state[self.alice.address].balance, 99)
        self.assertEqual(3, len(state_container.paginated_tx_hash.key_value))

        storage_key = state_container.paginated_tx_hash.generate_key(self.alice.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([tx.txhash], state_container.paginated_tx_hash.key_value[storage_key])

        storage_key = state_container.paginated_tx_hash.generate_key(self.bob.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([tx.txhash], state_container.paginated_tx_hash.key_value[storage_key])

        storage_key = state_container.paginated_tx_hash.generate_key(slave.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([tx.txhash], state_container.paginated_tx_hash.key_value[storage_key])

        self.assertEqual(2, len(state_container.tokens.data))
        self.assertIn((self.bob.address, tx.txhash), state_container.tokens.data)
        self.assertEqual(1000, state_container.tokens.data[(self.alice.address, tx.txhash)].balance)
        self.assertEqual(2000, state_container.tokens.data[(self.bob.address, tx.txhash)].balance)

        self.assertTrue(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(slave.address, 0))
        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))
        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.bob.address, 0))

    def test_revert_token_txn(self):
        """
        Same setup as in test_apply_state_changes(). This time though, the changes have already been applied,
        and we would like to roll them back.
        """
        initial_balances = [qrl_pb2.AddressAmount(address=self.alice.address, amount=1000),
                            qrl_pb2.AddressAmount(address=self.bob.address, amount=2000)]
        self.params["initial_balances"] = initial_balances

        tx = TokenTransaction.create(**self.params)
        tx.sign(self.alice)
        # Apply the changes!
        tx = TokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = dict(self.addresses_state)
        addresses_state[self.alice.address].pbdata.balance = 100
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)

        # After applying the Transaction, it should be as if Alice had never created the tokens in the first place.
        tx.apply(self.state, state_container)
        tx.revert(self.state, state_container)

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)

        self.assertEqual(2, len(state_container.paginated_tx_hash.key_value))

        storage_key = state_container.paginated_tx_hash.generate_key(self.alice.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([], state_container.paginated_tx_hash.key_value[storage_key])

        storage_key = state_container.paginated_tx_hash.generate_key(self.bob.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([], state_container.paginated_tx_hash.key_value[storage_key])

        self.assertEqual(2, len(state_container.tokens.data))
        self.assertIn((self.alice.address, tx.txhash), state_container.tokens.data)
        self.assertEqual(0, state_container.tokens.data[(self.alice.address, tx.txhash)].balance)
        self.assertIn((self.bob.address, tx.txhash), state_container.tokens.data)
        self.assertEqual(0, state_container.tokens.data[(self.bob.address, tx.txhash)].balance)
        self.assertEqual(0, addresses_state[self.alice.address].nonce)

    def test_revert_token_txn_owner_not_in_address_state(self):
        """
        In this case, Alice didn't give herself any tokens. How generous! She gave them all to Bob.
        But we want to revert this.
        """
        initial_balances = [qrl_pb2.AddressAmount(address=self.bob.address, amount=1000)]
        self.params["initial_balances"] = initial_balances

        tx = TokenTransaction.create(**self.params)
        tx.sign(self.alice)

        addresses_state = dict(self.addresses_state)
        addresses_state[self.alice.address].pbdata.balance = 100
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)

        state_container.paginated_bitfield.set_ots_key(addresses_state, self.alice.address, 0)

        self.assertTrue(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))
        tx.apply(self.state, state_container)
        tx.revert(self.state, state_container)
        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)

        self.assertEqual(2, len(state_container.paginated_tx_hash.key_value))

        storage_key = state_container.paginated_tx_hash.generate_key(self.alice.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([], state_container.paginated_tx_hash.key_value[storage_key])

        storage_key = state_container.paginated_tx_hash.generate_key(self.bob.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([], state_container.paginated_tx_hash.key_value[storage_key])

        self.assertEqual(1, len(state_container.tokens.data))
        self.assertIn((self.bob.address, tx.txhash), state_container.tokens.data)
        self.assertEqual(0, state_container.tokens.data[(self.bob.address, tx.txhash)].balance)

    def test_revert_token_txn_signed_by_slave_xmss(self):
        """
        Alice creates a token, gives herself and Bob some tokens.
        But she uses a XMSS slave to sign it.
        Can we undo it?
        """
        initial_balances = [qrl_pb2.AddressAmount(address=self.alice.address, amount=1000),
                            qrl_pb2.AddressAmount(address=self.bob.address, amount=2000)]
        slave = get_slave_xmss()
        self.params["initial_balances"] = initial_balances
        self.params["xmss_pk"] = slave.pk
        self.params["master_addr"] = self.alice.address
        tx = TokenTransaction.create(**self.params)
        tx.sign(slave)

        # Now that we have the Slave XMSS address, we should add it to AddressState so that apply_state_changes()
        # can do something with it
        addresses_state = dict(self.addresses_state)
        addresses_state[slave.address] = OptimizedAddressState.get_default(slave.address)
        addresses_state[slave.address].pbdata.nonce = 1

        # Also, update the AddressStates manually!
        addresses_state[self.alice.address].pbdata.balance = 100
        addresses_state[self.bob.address].pbdata.balance = 0
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        state_container.paginated_bitfield.set_ots_key(addresses_state, slave.address, 0)

        self.assertTrue(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(slave.address, 0))
        tx.apply(self.state, state_container)
        tx.revert(self.state, state_container)
        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(slave.address, 0))

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        self.assertEqual(3, len(state_container.paginated_tx_hash.key_value))

        storage_key = state_container.paginated_tx_hash.generate_key(self.alice.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([], state_container.paginated_tx_hash.key_value[storage_key])

        storage_key = state_container.paginated_tx_hash.generate_key(self.bob.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([], state_container.paginated_tx_hash.key_value[storage_key])

        storage_key = state_container.paginated_tx_hash.generate_key(slave.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([], state_container.paginated_tx_hash.key_value[storage_key])

        self.assertEqual(2, len(state_container.tokens.data))
        self.assertIn((self.alice.address, tx.txhash), state_container.tokens.data)
        self.assertEqual(0, state_container.tokens.data[(self.alice.address, tx.txhash)].balance)
        self.assertIn((self.bob.address, tx.txhash), state_container.tokens.data)
        self.assertEqual(0, state_container.tokens.data[(self.bob.address, tx.txhash)].balance)
        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))
        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.bob.address, 0))

    def test_validate_tx(self):
        initial_balances = [qrl_pb2.AddressAmount(address=self.alice.address, amount=1000),
                            qrl_pb2.AddressAmount(address=self.bob.address, amount=1000)]
        slave = get_slave_xmss()
        self.params["initial_balances"] = initial_balances
        self.params["xmss_pk"] = slave.pk
        self.params["master_addr"] = self.alice.address
        tx = TokenTransaction.create(**self.params)
        tx.sign(slave)

        self.assertTrue(tx.validate_or_raise())

        tx._data.transaction_hash = b'abc'

        # Should fail, as we have modified with invalid transaction_hash
        with self.assertRaises(ValueError):
            tx.validate_or_raise()



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from collections import OrderedDict

from pyqrllib.pyqrllib import bin2hstr
from twisted.internet import reactor

from qrl.core import config
from qrl.core.Block import Block
from qrl.core.ChainManager import ChainManager
from qrl.core.ESyncState import ESyncState
from qrl.core.Miner import Miner
from qrl.core.misc import ntp, logger


class SyncState:
    def __init__(self):
        self.state = ESyncState.unsynced
        self.epoch_diff = -1


class ConsensusMechanism(object):
    def __init__(self,
                 chain_manager: ChainManager):
        self.chain_manager = chain_manager


class POW(ConsensusMechanism):
    def __init__(self,
                 chain_manager: ChainManager,
                 p2p_factory,
                 sync_state: SyncState,
                 time_provider,
                 mining_address: bytes,
                 mining_thread_count):

        super().__init__(chain_manager)
        self.sync_state = sync_state
        self.time_provider = time_provider

        self.miner_toggler = False
        self.mining_address = mining_address

        self.p2p_factory = p2p_factory  # FIXME: Decouple from p2pFactory. Comms vs node logic
        self.p2p_factory.pow = self  # FIXME: Temporary hack to keep things working while refactoring

        self.miner = Miner(self.chain_manager,
                           self.pre_block_logic,
                           self.mining_address,
                           mining_thread_count)

        ########

        self.last_pow_cycle = 0
        self.last_bk_time = 0
        self.last_pb_time = 0
        self.suspend_mining_timestamp = 0

        self.future_blocks = OrderedDict()  # Keeps the list of future blocks, which has to be processed later

        self.epoch_diff = None

    ##################################################
    ##################################################
    ##################################################
    ##################################################

    def start(self):
        self.restart_monitor_bk(80)
        reactor.callLater(1, self.initialize_pow)

    def _handler_state_unsynced(self):
        self.miner.cancel()
        self.last_bk_time = ntp.getTime()
        self.restart_unsynced_logic()

    def _handler_state_syncing(self):
        self.last_pb_time = ntp.getTime()

    def _handler_state_synced(self):
        self.last_pow_cycle = ntp.getTime()
        last_block = self.chain_manager.last_block
        self._mine_next(last_block)

    def _handler_state_forked(self):
        pass

    def update_node_state(self, new_sync_state: ESyncState):
        self.sync_state.state = new_sync_state
        logger.info('Status changed to %s', self.sync_state.state)

        _mapping = {
            ESyncState.unsynced: self._handler_state_unsynced,
            ESyncState.syncing: self._handler_state_syncing,
            ESyncState.synced: self._handler_state_synced,
            ESyncState.forked: self._handler_state_forked,
        }

        _mapping[self.sync_state.state]()

    def stop_monitor_bk(self):
        try:
            reactor.monitor_bk.cancel()
        except Exception:  # No need to log this exception
            pass

    def restart_monitor_bk(self, delay: int):
        self.stop_monitor_bk()
        reactor.monitor_bk = reactor.callLater(delay, self.monitor_bk)

    def monitor_bk(self):
        # FIXME: Too many magic numbers / timing constants
        # FIXME: This is obsolete
        time_diff1 = ntp.getTime() - self.last_pow_cycle
        if 90 < time_diff1:
            if self.sync_state.state == ESyncState.unsynced:
                if ntp.getTime() - self.last_bk_time > 120:
                    self.last_pow_cycle = ntp.getTime()
                    logger.info(' POW cycle activated by monitor_bk() ')
                    self.update_node_state(ESyncState.synced)
                reactor.monitor_bk = reactor.callLater(60, self.monitor_bk)
                return

        time_diff2 = ntp.getTime() - self.last_pb_time
        if self.sync_state.state == ESyncState.syncing and time_diff2 > 60:
            self.update_node_state(ESyncState.unsynced)
            self.epoch_diff = -1

        reactor.monitor_bk = reactor.callLater(60, self.monitor_bk)

    def initialize_pow(self):
        reactor.callLater(0, self.update_node_state, ESyncState.synced)
        reactor.callLater(60, self.monitor_miner)

    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################

    def restart_unsynced_logic(self, delay=0):
        logger.info('Restarting unsynced logic in %s seconds', delay)
        try:
            reactor.unsynced_logic.cancel()
        except Exception:  # No need to log this exception
            pass

        reactor.unsynced_logic = reactor.callLater(delay, self.unsynced_logic)

    def unsynced_logic(self):
        if self.sync_state.state != ESyncState.synced:
            self.p2p_factory.broadcast_get_synced_state()
            reactor.request_peer_blockheight = reactor.callLater(0, self.p2p_factory.request_peer_blockheight)
            reactor.unsynced_logic = reactor.callLater(20, self.start_download)

    def start_download(self):
        # FIXME: Why PoW is downloading blocks?
        # add peers and their identity to requested list
        # FMBH
        if self.sync_state.state == ESyncState.synced:
            return

        logger.info('Checking Download..')

        if self.p2p_factory.num_connections == 0:
            logger.warning('No connected peers. Moving to synced state')
            self.update_node_state(ESyncState.synced)
            return

        self.update_node_state(ESyncState.syncing)
        logger.info('Initializing download from %s', self.chain_manager.height + 1)
        self.p2p_factory.randomize_block_fetch()

    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################

    def monitor_miner(self):
        reactor.callLater(15, self.monitor_miner)

        if not config.user.mining_enabled:
            return
        if not self.miner.isRunning() or self.miner_toggler:
            logger.debug('Mine next called by monitor_miner')
            self.miner_toggler = False
            self._mine_next(self.chain_manager.last_block)
        elif self.miner.solutionAvailable():
            self.miner_toggler = True
        else:
            self.miner_toggler = False

    def add_future_block(self, block):
        self.future_blocks[block.headerhash] = block
        if len(self.future_blocks) > config.dev.max_future_blocks_length:
            self.future_blocks.popitem(False)

    def process_future_blocks(self):
        keys = list(self.future_blocks.keys())
        for key in keys:
            block = self.future_blocks[key]
            dev_config = self.chain_manager.get_config_by_block_number(block.block_number)
            if block.is_future_block(dev_config):
                return
            self.pre_block_logic(block)
            del self.future_blocks[key]

    def pre_block_logic(self, block: Block):
        logger.debug('LOCK - TRY - pre_block_logic')
        with self.miner.lock:
            logger.debug('LOCK - LOCKED - pre_block_logic')

            if not block.validate(self.chain_manager, self.future_blocks):
                logger.warning('Block Validation failed for #%s %s', block.block_number, bin2hstr(block.headerhash))
                return False

            dev_config = self.chain_manager.get_config_by_block_number(block.block_number)
            if block.is_future_block(dev_config):
                delay = abs(block.timestamp - ntp.getTime()) + 1
                reactor.callLater(delay, self.process_future_blocks)
                self.add_future_block(block)
                return True

            logger.debug('Inside add_block')
            result = self.chain_manager.add_block(block)

            logger.debug('trigger_miner %s', self.chain_manager.trigger_miner)
            if self.chain_manager.trigger_miner:
                logger.debug('try last block')
                last_block = self.chain_manager.last_block
                logger.debug('got last block')
                self._mine_next(last_block)

            if not result:
                logger.debug('Block Rejected %s %s', block.block_number, bin2hstr(block.headerhash))
                return False

            reactor.callLater(0, self.broadcast_block, block)
        logger.debug('LOCK - RELEASE - pre_block_logic')

        return result

    def broadcast_block(self, block):
        if self.sync_state.state == ESyncState.synced:
            self.p2p_factory.broadcast_block(block)

    def isSynced(self, block_timestamp) -> bool:
        if block_timestamp + config.dev.minimum_minting_delay > ntp.getTime():
            self.update_node_state(ESyncState.synced)
            return True
        return False

    def _mine_next(self, parent_block):
        if ntp.getTime() < self.suspend_mining_timestamp:
            return

        if config.user.mining_enabled:
            logger.debug('try get_block_metadata')
            parent_metadata = self.chain_manager.get_block_metadata(parent_block.headerhash)
            logger.debug('try prepare_next_unmined_block_template')
            dev_config = self.chain_manager.get_config_by_block_number(parent_block.block_number + 1)
            self.miner.prepare_next_unmined_block_template(mining_address=self.mining_address,
                                                           tx_pool=self.chain_manager.tx_pool,
                                                           parent_block=parent_block,
                                                           parent_difficulty=parent_metadata.block_difficulty,
                                                           dev_config=dev_config)
            logger.info('Mining Block #%s', parent_block.block_number + 1)
            self.miner.start_mining(parent_block,
                                    parent_metadata.block_difficulty,
                                    dev_config)



from unittest import TestCase

from mock import Mock

from qrl.core import config
from qrl.core.Indexer import Indexer
from qrl.core.misc import logger
from qrl.core.State import State
from qrl.core.StateContainer import StateContainer
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.ChainManager import ChainManager
from qrl.core.txs.SlaveTransaction import SlaveTransaction
from qrl.generated.qrl_pb2 import SlaveMetadata
from tests.misc.helper import get_alice_xmss, get_slave_xmss, set_qrl_dir

logger.initialize_default()


class TestSlaveTransactionStateChanges(TestCase):
    def setUp(self):
        with set_qrl_dir('no_data'):
            self.state = State()

        self.alice = get_alice_xmss()
        self.slave = get_slave_xmss()
        alice_address_state = OptimizedAddressState.get_default(self.alice.address)
        alice_address_state.pbdata.balance = 100
        self.addresses_state = {
            self.alice.address: alice_address_state,
            self.slave.address: OptimizedAddressState.get_default(self.slave.address)
        }

        self.params = {
            "slave_pks": [self.slave.pk],
            "access_types": [0],
            "fee": 1,
            "xmss_pk": self.alice.pk
        }
        self.unused_chain_manager_mock = Mock(autospec=ChainManager, name='unused ChainManager')

    def test_apply_slave_txn(self):
        tx = SlaveTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = dict(self.addresses_state)
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        tx.apply(self.state, state_container)

        self.assertEqual(addresses_state[self.alice.address].balance, 99)

        storage_key = state_container.paginated_tx_hash.generate_key(self.alice.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([tx.txhash], state_container.paginated_tx_hash.key_value[storage_key])

        self.assertIn((tx.addr_from, tx.slave_pks[0]), state_container.slaves.data)
        data = state_container.slaves.data[(tx.addr_from, tx.slave_pks[0])]
        self.assertIsInstance(data, SlaveMetadata)
        self.assertEqual(tx.access_types[0], data.access_type)
        self.assertEqual(tx.txhash, data.tx_hash)

    def test_revert_slave_txn(self):
        tx = SlaveTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = dict(self.addresses_state)
        addresses_state[self.alice.address].pbdata.balance = 100
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        tx.apply(self.state, state_container)
        tx.revert(self.state, state_container)

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        storage_key = state_container.paginated_tx_hash.generate_key(self.alice.address, 1)
        self.assertIn(storage_key, state_container.paginated_tx_hash.key_value)
        self.assertEqual([], state_container.paginated_tx_hash.key_value[storage_key])

        self.assertIn((tx.addr_from, tx.slave_pks[0]), state_container.slaves.data)
        data = state_container.slaves.data[(tx.addr_from, tx.slave_pks[0])]
        self.assertIsInstance(data, SlaveMetadata)
        self.assertEqual(tx.access_types[0], data.access_type)
        self.assertEqual(tx.txhash, data.tx_hash)



from unittest import TestCase

from mock import patch, Mock

from qrl.core import config
from qrl.core.Indexer import Indexer
from qrl.core.misc import logger
from qrl.core.State import State
from qrl.core.StateContainer import StateContainer
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.txs.MessageTransaction import MessageTransaction
from tests.misc.helper import get_alice_xmss, get_bob_xmss, set_qrl_dir

logger.initialize_default()


@patch('qrl.core.txs.Transaction.logger')
class TestTransactionValidateSlave(TestCase):
    def setUp(self):
        with set_qrl_dir('no_data'):
            self.state = State()

        self.alice = get_alice_xmss()
        self.params = {
            "message_hash": b'Test Message',
            "addr_to": None,
            "fee": 1,
            "xmss_pk": self.alice.pk
        }
        self.m_addr_state = Mock(autospec=OptimizedAddressState, name='addr_state', balance=200)
        self.m_addr_from_pk_state = Mock(autospec=OptimizedAddressState, name='addr_from_pk_state')

    def test_validate_slave_valid(self, m_logger):
        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)
        result = tx.validate_slave(0)
        self.assertTrue(result)

    def test_validate_slave_master_addr_same_as_signing_addr(self, m_logger):
        self.params["master_addr"] = self.alice.address
        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)
        result = tx.validate_slave(None)
        self.assertFalse(result)

    def test_validate_slave_signing_xmss_state_has_no_slave_permissions_in_state(self, m_logger):
        bob = get_bob_xmss()
        # Let's say Alice is Bob's master.
        self.params["master_addr"] = self.alice.address
        self.params["xmss_pk"] = bob.pk

        # We need to add extra data to the mock AddressState.
        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)
        state_container = StateContainer(addresses_state=dict(),
                                         tokens=Indexer(b'token', None),
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        result = tx.validate_slave(state_container)
        self.assertFalse(result)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase
from mock import PropertyMock, patch
import random
from math import ceil

from qrl.core import config
from qrl.core.misc import logger
from qrl.core.State import State
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.PaginatedBitfield import PaginatedBitfield
from tests.misc.helper import get_alice_xmss, get_slave_xmss, set_qrl_dir

logger.initialize_default()

alice = get_alice_xmss()
slave = get_slave_xmss()


class TestPaginatedBitfield(TestCase):
    def setUp(self):
        pass

    def test_generate_bitfield_key(self):
        with set_qrl_dir('no_data'):
            state = State()
            paginated_bitfield = PaginatedBitfield(True, state._db)
            address = b'addr1'
            page = 1
            expected_key = b'bitfield_' + address + b'_' + page.to_bytes(8, byteorder='big', signed=False)
            found_key = paginated_bitfield.generate_bitfield_key(address, page)
            self.assertEqual(expected_key, found_key)

    def test_load_bitfield_and_ots_key_reuse(self):
        with set_qrl_dir('no_data'):
            state = State()
            alice_xmss = get_alice_xmss(4)
            paginated_bitfield = PaginatedBitfield(True, state._db)
            self.assertFalse(paginated_bitfield.load_bitfield_and_ots_key_reuse(alice_xmss.address,
                                                                                0))
            addresses_state = {
                alice_xmss.address: OptimizedAddressState.get_default(alice_xmss.address)
            }
            paginated_bitfield.set_ots_key(addresses_state, alice_xmss.address, 0)
            self.assertTrue(paginated_bitfield.load_bitfield_and_ots_key_reuse(alice_xmss.address,
                                                                               0))

    @patch('qrl.core.config.DevConfig.ots_tracking_per_page', new_callable=PropertyMock, return_value=1024)
    @patch('qrl.core.config.DevConfig.ots_bitfield_size', new_callable=PropertyMock)
    def test_ots_key_reuse(self, mock_ots_bitfield_size, mock_ots_tracking_per_page):
        """
        Randomly using OTS key
        :return:
        """
        with set_qrl_dir('no_data'):
            state = State()

            mock_ots_bitfield_size.return_value = ceil(config.dev.ots_tracking_per_page / 8)
            paginated_bitfield = PaginatedBitfield(True, state._db)

            alice_xmss = get_alice_xmss(12)
            address = alice_xmss.address
            address_state = OptimizedAddressState.get_default(address)
            addresses_state = {address: address_state}
            bitfield_data = paginated_bitfield.get_paginated_data(address, 1)
            self.assertFalse(paginated_bitfield.ots_key_reuse(bitfield_data, 0))

            paginated_bitfield.set_ots_key(addresses_state, address, 0)
            bitfield_data = paginated_bitfield.get_paginated_data(address, 1)
            # False, as bitfield has been set but has not been written to state.
            self.assertFalse(paginated_bitfield.ots_key_reuse(bitfield_data, 0))

            # Writing bitfield to the state.
            paginated_bitfield.put_addresses_bitfield(None)
            bitfield_data = paginated_bitfield.get_paginated_data(address, 1)
            self.assertTrue(paginated_bitfield.ots_key_reuse(bitfield_data, 0))

    @patch('qrl.core.config.DevConfig.ots_tracking_per_page', new_callable=PropertyMock, return_value=1024)
    @patch('qrl.core.config.DevConfig.ots_bitfield_size', new_callable=PropertyMock)
    def test_set_ots_key(self, mock_ots_bitfield_size, mock_ots_tracking_per_page):
        """
        Randomly using OTS key
        :return:
        """
        with set_qrl_dir('no_data'):
            state = State()
            mock_ots_bitfield_size.return_value = ceil(config.dev.ots_tracking_per_page / 8)

            alice_xmss = get_alice_xmss(12)
            address = alice_xmss.address
            address_state = OptimizedAddressState.get_default(address)
            addresses_state = {address: address_state}
            paginated_bitfield = PaginatedBitfield(True, state._db)
            paginated_bitfield.update_used_page_in_address_state(address, addresses_state, 1)
            self.assertEqual(address_state.ots_bitfield_used_page, 0)

            ots_indexes = list(range(3072, 2 ** alice_xmss.height))
            random.shuffle(ots_indexes)
            for i in ots_indexes:
                paginated_bitfield.set_ots_key(addresses_state, address, i)
                self.assertEqual(address_state.ots_bitfield_used_page, 0)

            ots_indexes = list(range(2048, 3072))
            random.shuffle(ots_indexes)
            for i in ots_indexes:
                paginated_bitfield.set_ots_key(addresses_state, address, i)
                self.assertEqual(address_state.ots_bitfield_used_page, 0)

            ots_indexes = list(range(1024, 2048))
            random.shuffle(ots_indexes)
            for i in ots_indexes:
                paginated_bitfield.set_ots_key(addresses_state, address, i)
                self.assertEqual(address_state.ots_bitfield_used_page, 0)

            ots_indexes = list(range(0, 1024))
            random.shuffle(ots_indexes)
            for i in ots_indexes:
                paginated_bitfield.set_ots_key(addresses_state, address, i)
                if i == ots_indexes[-1]:
                    self.assertEqual(address_state.ots_bitfield_used_page, 4)

            self.assertEqual(address_state.ots_bitfield_used_page, 4)

    @patch('qrl.core.config.DevConfig.ots_tracking_per_page', new_callable=PropertyMock, return_value=1024)
    @patch('qrl.core.config.DevConfig.ots_bitfield_size', new_callable=PropertyMock)
    def test_set_ots_key2(self, mock_ots_bitfield_size, mock_ots_tracking_per_page):
        """
        Randomly using OTS key
        :return:
        """
        with set_qrl_dir('no_data'):
            state = State()
            mock_ots_bitfield_size.return_value = ceil(config.dev.ots_tracking_per_page / 8)

            alice_xmss = get_alice_xmss(12)
            address = alice_xmss.address
            address_state = OptimizedAddressState.get_default(address)
            addresses_state = {address: address_state}
            paginated_bitfield = PaginatedBitfield(True, state._db)
            paginated_bitfield.update_used_page_in_address_state(address, addresses_state, 1)
            self.assertEqual(address_state.ots_bitfield_used_page, 0)

            ots_indexes = list(range(0, 2**alice_xmss.height))
            random.shuffle(ots_indexes)
            for i in ots_indexes:
                paginated_bitfield.set_ots_key(addresses_state, address, i)
                if i == ots_indexes[-1]:
                    self.assertEqual(address_state.ots_bitfield_used_page, 4)

            self.assertEqual(address_state.ots_bitfield_used_page, 4)

    @patch('qrl.core.config.DevConfig.ots_tracking_per_page', new_callable=PropertyMock, return_value=1024)
    @patch('qrl.core.config.DevConfig.ots_bitfield_size', new_callable=PropertyMock)
    def test_unset_ots_key(self, mock_ots_bitfield_size, mock_ots_tracking_per_page):
        """
        Randomly using OTS key
        :return:
        """
        with set_qrl_dir('no_data'):
            state = State()
            mock_ots_bitfield_size.return_value = ceil(config.dev.ots_tracking_per_page / 8)

            alice_xmss = get_alice_xmss(12)
            address = alice_xmss.address
            address_state = OptimizedAddressState.get_default(address)
            addresses_state = {address: address_state}
            paginated_bitfield = PaginatedBitfield(True, state._db)
            paginated_bitfield.update_used_page_in_address_state(address, addresses_state, 1)
            self.assertEqual(address_state.ots_bitfield_used_page, 0)

            ots_indexes = list(range(3072, 2 ** alice_xmss.height))
            random.shuffle(ots_indexes)
            for i in ots_indexes:
                paginated_bitfield.set_ots_key(addresses_state, address, i)
                self.assertEqual(address_state.ots_bitfield_used_page, 0)

            ots_indexes = list(range(2048, 3072))
            random.shuffle(ots_indexes)
            for i in ots_indexes:
                paginated_bitfield.set_ots_key(addresses_state, address, i)
                self.assertEqual(address_state.ots_bitfield_used_page, 0)

            ots_indexes = list(range(1024, 2048))
            random.shuffle(ots_indexes)
            for i in ots_indexes:
                paginated_bitfield.set_ots_key(addresses_state, address, i)
                self.assertEqual(address_state.ots_bitfield_used_page, 0)

            ots_indexes = list(range(0, 1024))
            random.shuffle(ots_indexes)
            for i in ots_indexes:
                paginated_bitfield.set_ots_key(addresses_state, address, i)
                if i == ots_indexes[-1]:
                    self.assertEqual(address_state.ots_bitfield_used_page, 4)
                else:
                    self.assertEqual(address_state.ots_bitfield_used_page, 0)

            self.assertEqual(address_state.ots_bitfield_used_page, 4)

            paginated_bitfield.unset_ots_key(addresses_state, address, 2049)
            self.assertEqual(address_state.ots_bitfield_used_page, 2)

            paginated_bitfield.unset_ots_key(addresses_state, address, 1023)
            self.assertEqual(address_state.ots_bitfield_used_page, 0)

            paginated_bitfield.set_ots_key(addresses_state, address, 1023)
            self.assertEqual(address_state.ots_bitfield_used_page, 2)

    @patch('qrl.core.config.DevConfig.ots_tracking_per_page', new_callable=PropertyMock, return_value=1024)
    @patch('qrl.core.config.DevConfig.ots_bitfield_size', new_callable=PropertyMock)
    def test_unset_ots_key2(self, mock_ots_bitfield_size, mock_ots_tracking_per_page):
        """
        Randomly using OTS key
        :return:
        """
        with set_qrl_dir('no_data'):
            state = State()
            mock_ots_bitfield_size.return_value = ceil(config.dev.ots_tracking_per_page / 8)

            alice_xmss = get_alice_xmss(12)
            address = alice_xmss.address
            address_state = OptimizedAddressState.get_default(address)
            addresses_state = {address: address_state}
            paginated_bitfield = PaginatedBitfield(True, state._db)
            paginated_bitfield.update_used_page_in_address_state(address, addresses_state, 1)
            self.assertEqual(address_state.ots_bitfield_used_page, 0)

            ots_indexes = list(range(1, 2**alice_xmss.height))
            random.shuffle(ots_indexes)
            ots_indexes.append(0)
            for i in ots_indexes:
                paginated_bitfield.set_ots_key(addresses_state, address, i)
                if i == ots_indexes[-1]:
                    self.assertEqual(address_state.ots_bitfield_used_page, 4)
                else:
                    self.assertEqual(address_state.ots_bitfield_used_page, 0)

            self.assertEqual(address_state.ots_bitfield_used_page, 4)

            paginated_bitfield.unset_ots_key(addresses_state, address, 4095)
            self.assertEqual(address_state.ots_bitfield_used_page, 3)

            paginated_bitfield.unset_ots_key(addresses_state, address, 3000)
            self.assertEqual(address_state.ots_bitfield_used_page, 2)

            paginated_bitfield.unset_ots_key(addresses_state, address, 2000)
            self.assertEqual(address_state.ots_bitfield_used_page, 1)

            paginated_bitfield.unset_ots_key(addresses_state, address, 1000)
            self.assertEqual(address_state.ots_bitfield_used_page, 0)

            paginated_bitfield.set_ots_key(addresses_state, address, 1000)
            self.assertEqual(address_state.ots_bitfield_used_page, 1)

            paginated_bitfield.set_ots_key(addresses_state, address, 2000)
            self.assertEqual(address_state.ots_bitfield_used_page, 2)

            paginated_bitfield.set_ots_key(addresses_state, address, 3000)
            self.assertEqual(address_state.ots_bitfield_used_page, 3)

            paginated_bitfield.set_ots_key(addresses_state, address, 4095)
            self.assertEqual(address_state.ots_bitfield_used_page, 4)

    @patch('qrl.core.config.DevConfig.ots_tracking_per_page', new_callable=PropertyMock, return_value=1024)
    @patch('qrl.core.config.DevConfig.ots_bitfield_size', new_callable=PropertyMock)
    def test_unset_ots_key3(self, mock_ots_bitfield_size, mock_ots_tracking_per_page):
        """
        Features Tested
        - Sequentially marking OTS indexes as used
        - Sequentially marking OTS indexes as unused
        - ots_bitfield_used_page value with each OTS index being used

        Expectation
        - The ots_bitfield_used_page must increase by 1 for every sequential 1024 (ots_tracking_per_page) ots indexes
          marked as used

        :param mock_ots_bitfield_size:
        :param mock_ots_tracking_per_page:
        :return:
        """
        with set_qrl_dir('no_data'):
            state = State()
            mock_ots_bitfield_size.return_value = ceil(config.dev.ots_tracking_per_page / 8)

            alice_xmss = get_alice_xmss(12)
            address = alice_xmss.address
            address_state = OptimizedAddressState.get_default(address)
            addresses_state = {address: address_state}
            paginated_bitfield = PaginatedBitfield(True, state._db)
            paginated_bitfield.update_used_page_in_address_state(address, addresses_state, 1)
            self.assertEqual(address_state.ots_bitfield_used_page, 0)

            total_ots = 2 ** alice_xmss.height
            for i in range(0, total_ots + 1):
                paginated_bitfield.set_ots_key(addresses_state, address, i)
                self.assertEqual(address_state.ots_bitfield_used_page, (i + 1) // config.dev.ots_tracking_per_page)

            self.assertEqual(address_state.ots_bitfield_used_page,
                             total_ots // config.dev.ots_tracking_per_page)
            self.assertEqual(total_ots // config.dev.ots_tracking_per_page, 4)

            paginated_bitfield.unset_ots_key(addresses_state, address, total_ots - 1)
            self.assertEqual(address_state.ots_bitfield_used_page, 3)

            for i in range(total_ots - 2, -1, -1):
                paginated_bitfield.unset_ots_key(addresses_state, address, i)
                self.assertEqual(address_state.ots_bitfield_used_page, i // config.dev.ots_tracking_per_page)



from qrl.core.config import DevConfig
from qrl.core.Indexer import Indexer
from qrl.core.PaginatedData import PaginatedData
from qrl.core.PaginatedBitfield import PaginatedBitfield
from qrl.core.misc import db, logger


class StateContainer:
    def __init__(self,
                 addresses_state: dict,
                 tokens: Indexer,
                 slaves: Indexer,
                 lattice_pk: Indexer,
                 multi_sig_spend_txs: dict,
                 votes_stats: dict,
                 block_number: int,
                 total_coin_supply: int,
                 current_dev_config: DevConfig,
                 write_access: bool,
                 my_db: db,
                 batch):
        self.delete_keys = set()
        self.paginated_bitfield = PaginatedBitfield(write_access, my_db)
        self.paginated_tx_hash = PaginatedData(b'p_tx_hash', write_access, my_db)
        self.paginated_tokens_hash = PaginatedData(b'p_tokens', write_access, my_db)
        self.paginated_slaves_hash = PaginatedData(b'p_slaves', write_access, my_db)
        self.paginated_lattice_pk = PaginatedData(b'p_lattice_pk', write_access, my_db)
        self.paginated_multisig_address = PaginatedData(b'p_multisig_address', write_access, my_db)
        self.paginated_multi_sig_spend = PaginatedData(b'p_multi_sig_spend', write_access, my_db)
        self.paginated_inbox_message = PaginatedData(b'p_inbox_message', write_access, my_db)

        self.addresses_state = addresses_state

        self.tokens = tokens
        self.slaves = slaves
        self.lattice_pk = lattice_pk
        self.multi_sig_spend_txs = multi_sig_spend_txs
        self.votes_stats = votes_stats
        self.block_number = block_number  # Block number that is being processed
        self.batch = batch
        self.db = my_db
        self.current_dev_config = current_dev_config

        # Keeps track of last update so that it can be reverted
        self.last_addresses_state = dict()
        self.last_tokens = Indexer(b'token', self.db)
        self.last_slaves = Indexer(b'slave', self.db)
        self.last_lattice_pk = Indexer(b'lattice_pk', self.db)
        self.last_multi_sig_spend_txs = dict()
        self.last_votes_stats = dict()

        self._total_coin_supply = total_coin_supply  # TODO: Coinbase transaction of current block is not included

    @property
    def total_coin_supply(self):
        return self._total_coin_supply

    @staticmethod
    def _copy_key_value(src: dict, dst: dict) -> bool:
        for key in src:
            if key in dst:
                StateContainer._revert_copy_key_value(src, dst, False)
                return False
            dst[key] = src[key]

        return True

    @staticmethod
    def _revert_copy_key_value(src: dict, dst: dict, error_if_key_not_found: bool) -> bool:
        for key in src:
            if key in dst:
                del dst[key]
            elif error_if_key_not_found:
                logger.error("Key %s not found while reverting key from state_container", key)
                return False

        return True

    def update(self,
               addresses_state: dict,
               tokens: Indexer,
               slaves: Indexer,
               lattice_pk: Indexer,
               multi_sig_spend_txs: dict,
               votes_stats: dict) -> bool:

        self.last_addresses_state = addresses_state
        self.last_tokens = tokens
        self.last_slaves = slaves
        self.last_lattice_pk = lattice_pk
        self.last_multi_sig_spend_txs = multi_sig_spend_txs
        self.last_votes_stats = votes_stats

        if not self._copy_key_value(addresses_state, self.addresses_state):
            logger.error("Error updating addresses_state in state_container")
            return False

        if not self._copy_key_value(tokens.data, self.tokens.data):
            logger.error("Error updating tokens in state_container")
            return False

        if not self._copy_key_value(slaves.data, self.slaves.data):
            logger.error("Error updating slaves in state_container")
            return False

        if not self._copy_key_value(lattice_pk.data, self.lattice_pk.data):
            logger.error("Error updating lattice_pk in state_container")
            return False

        if not self._copy_key_value(multi_sig_spend_txs, self.multi_sig_spend_txs):
            logger.error("Error updating multi_sig_spend_txs in state_container")
            return False

        if not self._copy_key_value(votes_stats, self.votes_stats):
            logger.error("Error updating votes_stats in state_container")
            return False

        return True

    def revert_update(self) -> bool:
        if not self._revert_copy_key_value(self.last_addresses_state, self.addresses_state, True):
            logger.error("Error reverting last_addresses_state from state_container")
            return False
        self.last_addresses_state = dict()

        if not self._revert_copy_key_value(self.last_tokens.data, self.tokens.data, True):
            logger.error("Error reverting last_tokens from state_container")
            return False
        self.last_tokens = Indexer(b'token', self.db)

        if not self._revert_copy_key_value(self.last_slaves.data, self.slaves.data, True):
            logger.error("Error reverting last_slaves from state_container")
            return False
        self.last_slaves = Indexer(b'slave', self.db)

        if not self._revert_copy_key_value(self.last_lattice_pk.data, self.lattice_pk.data, True):
            logger.error("Error reverting last_lattice_pk from state_container")
            return False
        self.last_lattice_pk = Indexer(b'lattice_pk', self.db)

        if not self._revert_copy_key_value(self.last_multi_sig_spend_txs, self.multi_sig_spend_txs, True):
            logger.error("Error reverting last_multi_sig_spend_txs from state_container")
            return False
        self.last_multi_sig_spend_txs = dict()

        if not self._revert_copy_key_value(self.last_votes_stats, self.votes_stats, True):
            logger.error("Error reverting last_votes_stats from state_container")
            return False
        self.last_votes_stats = dict()

        return True



from unittest import TestCase

from qrl.core import config
from qrl.core.Indexer import Indexer
from qrl.core.State import State
from qrl.core.StateContainer import StateContainer
from qrl.core.misc import logger
from qrl.core.VoteStats import VoteStats
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.MultiSigAddressState import MultiSigAddressState
from qrl.core.txs.multisig.MultiSigSpend import MultiSigSpend
from qrl.core.txs.multisig.MultiSigVote import MultiSigVote
from tests.misc.helper import get_alice_xmss, get_bob_xmss, set_qrl_dir

logger.initialize_default()


class TestVoteStats(TestCase):
    def setUp(self) -> None:
        with set_qrl_dir('no_data'):
            self.state = State()

        self.alice = get_alice_xmss()
        self.bob = get_bob_xmss()
        self.random = get_alice_xmss(4)
        self.random_signer = get_bob_xmss(4)
        self.signatories = [self.alice.address, self.bob.address, self.random.address]
        self.weights = [20, 30, 10]
        self.threshold = 30

    def test_apply_and_put(self):
        """
        Test execution of multisig spend txn, when threshold is not met.
        :return:
        """
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx = MultiSigVote.create(shared_key=spend_tx.txhash,
                                 unvote=False,
                                 fee=5,
                                 xmss_pk=self.alice.pk)
        tx.sign(self.alice)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=100,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=5,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', self.state._db),
                                         slaves=Indexer(b'slave', self.state._db),
                                         lattice_pk=Indexer(b'lattice_pk', self.state._db),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=10,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        self.assertTrue(tx.apply(self.state, state_container))

        VoteStats.put_all(self.state, state_container)
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 100)
        # Expected balance is 0, as 5 quanta paid in MultiSigVote txn as a fee
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 0)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

    def test_revert_and_put(self):
        """
        Test execution of multisig spend txn, when threshold is not met and then reverting the txn.
        :return:
        """
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx = MultiSigVote.create(shared_key=spend_tx.txhash,
                                 unvote=False,
                                 fee=5,
                                 xmss_pk=self.alice.pk)
        tx.sign(self.alice)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=100,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=5,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', self.state._db),
                                         slaves=Indexer(b'slave', self.state._db),
                                         lattice_pk=Indexer(b'lattice_pk', self.state._db),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=10,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        self.assertTrue(tx.apply(self.state, state_container))

        VoteStats.put_all(self.state, state_container)
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 100)
        # Expected balance is 0, as 5 quanta paid in MultiSigVote txn as a fee
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 0)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

        VoteStats.revert_all(self.state, state_container)
        self.assertTrue(tx.revert(self.state, state_container))

        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 100)
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 5)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

    def test_apply_and_put2(self):
        """
        Test execution of multi sig txn, when threshold is met.
        :return:
        """
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx1 = MultiSigVote.create(shared_key=spend_tx.txhash,
                                  unvote=False,
                                  fee=5,
                                  xmss_pk=self.alice.pk)
        tx1.sign(self.alice)

        tx2 = MultiSigVote.create(shared_key=spend_tx.txhash,
                                  unvote=False,
                                  fee=4,
                                  xmss_pk=self.bob.pk)
        tx2.sign(self.bob)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        bob_address_state = OptimizedAddressState.get_default(address=self.bob.address)
        bob_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=100,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=5,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            self.bob.address: bob_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', self.state._db),
                                         slaves=Indexer(b'slave', self.state._db),
                                         lattice_pk=Indexer(b'lattice_pk', self.state._db),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=10,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        self.assertTrue(tx1.apply(self.state, state_container))
        self.assertTrue(tx2.apply(self.state, state_container))

        VoteStats.put_all(self.state, state_container)
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 0)
        # Expected balance is 100, as 5 quanta paid in MultiSigVote txn as a fee
        # and 100 quanta received by multi sig spend txn
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 100)
        # Expected balance is 1, as 4 quanta paid in MultiSigVote txn as a fee
        self.assertEqual(state_container.addresses_state[self.bob.address].balance, 1)
        self.assertTrue(vote_stats[spend_tx.txhash].executed)

    def test_revert_and_put2(self):
        """
        Test execution of multi sig txn, when threshold is met and then reverting the txn.
        :return:
        """
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx1 = MultiSigVote.create(shared_key=spend_tx.txhash,
                                  unvote=False,
                                  fee=5,
                                  xmss_pk=self.alice.pk)
        tx1.sign(self.alice)

        tx2 = MultiSigVote.create(shared_key=spend_tx.txhash,
                                  unvote=False,
                                  fee=4,
                                  xmss_pk=self.bob.pk)
        tx2.sign(self.bob)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        bob_address_state = OptimizedAddressState.get_default(address=self.bob.address)
        bob_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=100,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=5,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            self.bob.address: bob_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', self.state._db),
                                         slaves=Indexer(b'slave', self.state._db),
                                         lattice_pk=Indexer(b'lattice_pk', self.state._db),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=10,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        self.assertTrue(tx1.apply(self.state, state_container))
        self.assertTrue(tx2.apply(self.state, state_container))

        VoteStats.put_all(self.state, state_container)
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 0)
        # Expected balance is 100, as 5 quanta paid in MultiSigVote txn as a fee
        # and 100 quanta received by multi sig spend txn
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 100)
        # Expected balance is 1, as 4 quanta paid in MultiSigVote txn as a fee
        self.assertEqual(state_container.addresses_state[self.bob.address].balance, 1)
        self.assertTrue(vote_stats[spend_tx.txhash].executed)

        VoteStats.revert_all(self.state, state_container)
        self.assertTrue(tx1.revert(self.state, state_container))
        self.assertTrue(tx2.revert(self.state, state_container))

        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 100)
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 5)
        self.assertEqual(state_container.addresses_state[self.bob.address].balance, 5)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

    def test_apply_and_put3(self):
        """
        Test execution of multisig spend txn, when threshold is met by a single txn.
        :return:
        """
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx = MultiSigVote.create(shared_key=spend_tx.txhash,
                                 unvote=False,
                                 fee=5,
                                 xmss_pk=self.alice.pk)
        tx.sign(self.alice)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=100,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=4,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', self.state._db),
                                         slaves=Indexer(b'slave', self.state._db),
                                         lattice_pk=Indexer(b'lattice_pk', self.state._db),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=10,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        self.assertTrue(tx.apply(self.state, state_container))

        VoteStats.put_all(self.state, state_container)
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 0)
        # Expected balance is 100, as 5 quanta paid in MultiSigVote txn as a fee
        # and 100 quanta received by multi sig spend txn
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 100)
        self.assertTrue(vote_stats[spend_tx.txhash].executed)

    def test_revert_and_put3(self):
        """
        Test execution of multisig spend txn, when threshold is met by a single txn and then reverting the txn.
        :return:
        """
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx = MultiSigVote.create(shared_key=spend_tx.txhash,
                                 unvote=False,
                                 fee=5,
                                 xmss_pk=self.alice.pk)
        tx.sign(self.alice)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=100,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=4,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', self.state._db),
                                         slaves=Indexer(b'slave', self.state._db),
                                         lattice_pk=Indexer(b'lattice_pk', self.state._db),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=10,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        self.assertTrue(tx.apply(self.state, state_container))

        VoteStats.put_all(self.state, state_container)
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 0)
        # Expected balance is 100, as 5 quanta paid in MultiSigVote txn as a fee
        # and 100 quanta received by multi sig spend txn
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 100)
        self.assertTrue(vote_stats[spend_tx.txhash].executed)

        VoteStats.revert_all(self.state, state_container)
        self.assertTrue(tx.revert(self.state, state_container))
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 100)
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 5)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

    def test_apply_and_put4(self):
        """
        Test execution of multisig spend txn, when threshold is met but multi_sig address doesn't have
        sufficient balance.
        :return:
        """
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx = MultiSigVote.create(shared_key=spend_tx.txhash,
                                 unvote=False,
                                 fee=5,
                                 xmss_pk=self.alice.pk)
        tx.sign(self.alice)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=90,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=4,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', self.state._db),
                                         slaves=Indexer(b'slave', self.state._db),
                                         lattice_pk=Indexer(b'lattice_pk', self.state._db),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=10,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        self.assertTrue(tx.apply(self.state, state_container))

        VoteStats.put_all(self.state, state_container)
        # Expected balance is 90, as multi_sig_spend txn is not executed due to insufficient balance
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 90)
        # Expected balance is 0, as 5 quanta paid in MultiSigVote txn as a fee
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 0)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

    def test_revert_and_put4(self):
        """
        Test execution of multisig spend txn, when threshold is met but multi_sig address doesn't have
        sufficient balance and then reverting the txn.
        :return:
        """
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx = MultiSigVote.create(shared_key=spend_tx.txhash,
                                 unvote=False,
                                 fee=5,
                                 xmss_pk=self.alice.pk)
        tx.sign(self.alice)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=90,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=4,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', self.state._db),
                                         slaves=Indexer(b'slave', self.state._db),
                                         lattice_pk=Indexer(b'lattice_pk', self.state._db),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=10,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        self.assertTrue(tx.apply(self.state, state_container))

        VoteStats.put_all(self.state, state_container)
        # Expected balance is 90, as multi_sig_spend txn is not executed due to insufficient balance
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 90)
        # Expected balance is 0, as 5 quanta paid in MultiSigVote txn as a fee
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 0)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

        VoteStats.revert_all(self.state, state_container)
        self.assertTrue(tx.revert(self.state, state_container))

        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 90)
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 5)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

    def test_apply_and_put5(self):
        """
        Test execution of multisig spend txn, when threshold is met but multi_sig address doesn't have
        sufficient balance, later somehow multi sig address gets sufficient balance, before expiry.
        :return:
        """
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx1 = MultiSigVote.create(shared_key=spend_tx.txhash,
                                  unvote=False,
                                  fee=5,
                                  xmss_pk=self.alice.pk)
        tx1.sign(self.alice)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=90,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=4,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', self.state._db),
                                         slaves=Indexer(b'slave', self.state._db),
                                         lattice_pk=Indexer(b'lattice_pk', self.state._db),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=10,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        self.assertTrue(tx1.apply(self.state, state_container))

        VoteStats.put_all(self.state, state_container)
        # Expected balance is 90, as multi_sig_spend txn is not executed due to insufficient balance
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 90)
        # Expected balance is 0, as 5 quanta paid in MultiSigVote txn as a fee
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 0)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

        multi_sig_address_state.pbdata.balance = 100
        tx2 = MultiSigVote.create(shared_key=spend_tx.txhash,
                                  unvote=True,
                                  fee=0,
                                  xmss_pk=self.alice.pk)
        tx2.sign(self.alice)

        self.assertTrue(tx2.apply(self.state, state_container))
        VoteStats.put_all(self.state, state_container)
        # Expected balance is 100
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 100)
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 0)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

        tx3 = MultiSigVote.create(shared_key=spend_tx.txhash,
                                  unvote=False,
                                  fee=0,
                                  xmss_pk=self.alice.pk)
        tx3.sign(self.alice)
        self.assertTrue(tx3.apply(self.state, state_container))
        VoteStats.put_all(self.state, state_container)
        # Expected balance is 0, as multi_sig_spend txn has been executed
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 0)
        # Expected balance is 100, as 5 quanta paid in MultiSigVote txn as a fee
        # and 100 quanta received by multi sig spend txn
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 100)
        self.assertTrue(vote_stats[spend_tx.txhash].executed)

    def test_revert_and_put5(self):
        """
        Test execution of multisig spend txn, when threshold is met but multi_sig address doesn't have
        sufficient balance, later somehow multi sig address gets sufficient balance, before expiry.
        Then fork recovery is called.
        :return:
        """
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx1 = MultiSigVote.create(shared_key=spend_tx.txhash,
                                  unvote=False,
                                  fee=5,
                                  xmss_pk=self.alice.pk)
        tx1.sign(self.alice)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=90,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=4,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', self.state._db),
                                         slaves=Indexer(b'slave', self.state._db),
                                         lattice_pk=Indexer(b'lattice_pk', self.state._db),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=10,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        self.assertTrue(tx1.apply(self.state, state_container))

        VoteStats.put_all(self.state, state_container)
        # Expected balance is 90, as multi_sig_spend txn is not executed due to insufficient balance
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 90)
        # Expected balance is 0, as 5 quanta paid in MultiSigVote txn as a fee
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 0)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

        multi_sig_address_state.pbdata.balance = 100
        tx2 = MultiSigVote.create(shared_key=spend_tx.txhash,
                                  unvote=True,
                                  fee=0,
                                  xmss_pk=self.alice.pk)
        tx2.sign(self.alice)

        self.assertTrue(tx2.apply(self.state, state_container))
        VoteStats.put_all(self.state, state_container)
        # Expected balance is 100
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 100)
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 0)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

        tx3 = MultiSigVote.create(shared_key=spend_tx.txhash,
                                  unvote=False,
                                  fee=0,
                                  xmss_pk=self.alice.pk)
        tx3.sign(self.alice)
        self.assertTrue(tx3.apply(self.state, state_container))
        VoteStats.put_all(self.state, state_container)
        # Expected balance is 0, as multi_sig_spend txn has been executed
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 0)
        # Expected balance is 100, as 5 quanta paid in MultiSigVote txn as a fee
        # and 100 quanta received by multi sig spend txn
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 100)
        self.assertTrue(vote_stats[spend_tx.txhash].executed)

        VoteStats.revert_all(self.state, state_container)
        self.assertTrue(tx3.revert(self.state, state_container))
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 100)
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 0)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

        multi_sig_address_state.pbdata.balance = 90

        VoteStats.revert_all(self.state, state_container)
        self.assertTrue(tx2.revert(self.state, state_container))
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 90)
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 0)

        VoteStats.revert_all(self.state, state_container)
        self.assertTrue(tx1.revert(self.state, state_container))
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 90)
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 5)

    def test_apply_and_put6(self):
        """
        Test execution of multi sig txn, when threshold is met but after the block expiry.
        :return:
        """
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx1 = MultiSigVote.create(shared_key=spend_tx.txhash,
                                  unvote=False,
                                  fee=5,
                                  xmss_pk=self.alice.pk)
        tx1.sign(self.alice)

        tx2 = MultiSigVote.create(shared_key=spend_tx.txhash,
                                  unvote=False,
                                  fee=4,
                                  xmss_pk=self.bob.pk)
        tx2.sign(self.bob)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        bob_address_state = OptimizedAddressState.get_default(address=self.bob.address)
        bob_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=100,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=5,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            self.bob.address: bob_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', self.state._db),
                                         slaves=Indexer(b'slave', self.state._db),
                                         lattice_pk=Indexer(b'lattice_pk', self.state._db),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=15001,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        self.assertFalse(tx1.apply(self.state, state_container))
        self.assertFalse(tx2.apply(self.state, state_container))

        VoteStats.put_all(self.state, state_container)
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 100)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

    def test_revert_and_put6(self):
        """
        Test execution of multi sig txn, when threshold is met but after the block expiry.
        :return:
        """
        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(b'')
        spend_tx = MultiSigSpend.create(multi_sig_address=multi_sig_address,
                                        addrs_to=[self.alice.address],
                                        amounts=[100],
                                        expiry_block_number=15000,
                                        fee=0,
                                        xmss_pk=self.alice.pk)
        spend_tx.sign(self.alice)
        tx1 = MultiSigVote.create(shared_key=spend_tx.txhash,
                                  unvote=False,
                                  fee=5,
                                  xmss_pk=self.alice.pk)
        tx1.sign(self.alice)

        tx2 = MultiSigVote.create(shared_key=spend_tx.txhash,
                                  unvote=False,
                                  fee=4,
                                  xmss_pk=self.bob.pk)
        tx2.sign(self.bob)

        alice_address_state = OptimizedAddressState.get_default(address=self.alice.address)
        alice_address_state.pbdata.balance = 5
        bob_address_state = OptimizedAddressState.get_default(address=self.bob.address)
        bob_address_state.pbdata.balance = 5
        multi_sig_address_state = MultiSigAddressState.create(creation_tx_hash=b'',
                                                              balance=100,
                                                              signatories=[self.alice.address, self.bob.address],
                                                              weights=[4, 6],
                                                              threshold=5,
                                                              transaction_hash_count=0)
        addresses_state = {
            self.alice.address: alice_address_state,
            self.bob.address: bob_address_state,
            multi_sig_address: multi_sig_address_state,
        }
        vote_stats = {
            spend_tx.txhash: VoteStats.create(multi_sig_address=multi_sig_address,
                                              shared_key=spend_tx.txhash,
                                              signatories=multi_sig_address_state.signatories,
                                              expiry_block_number=spend_tx.expiry_block_number),
        }
        multi_sig_spend_txs = {
            spend_tx.txhash: spend_tx,
        }
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=Indexer(b'token', self.state._db),
                                         slaves=Indexer(b'slave', self.state._db),
                                         lattice_pk=Indexer(b'lattice_pk', self.state._db),
                                         multi_sig_spend_txs=multi_sig_spend_txs,
                                         votes_stats=vote_stats,
                                         block_number=15001,
                                         total_coin_supply=100,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        self.assertFalse(tx1.apply(self.state, state_container))
        self.assertFalse(tx2.apply(self.state, state_container))

        VoteStats.put_all(self.state, state_container)
        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 100)
        self.assertEqual(state_container.addresses_state[self.alice.address].balance, 0)
        # Expected balance is 1, as 4 quanta paid in MultiSigVote txn as a fee
        self.assertEqual(state_container.addresses_state[self.bob.address].balance, 1)
        self.assertFalse(vote_stats[spend_tx.txhash].executed)

        VoteStats.revert_all(self.state, state_container)
        self.assertFalse(tx2.revert(self.state, state_container))
        self.assertFalse(tx1.revert(self.state, state_container))

        self.assertEqual(state_container.addresses_state[multi_sig_address].balance, 100)



from unittest import TestCase

import simplejson as json
from mock import patch, PropertyMock
from pyqrllib.pyqrllib import bin2hstr

from qrl.core import config
from qrl.core.Indexer import Indexer
from qrl.core.misc import logger
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.State import State
from qrl.core.StateContainer import StateContainer
from qrl.core.txs.Transaction import Transaction
from qrl.core.txs.TransferTokenTransaction import TransferTokenTransaction
from qrl.generated import qrl_pb2
from tests.core.txs.testdata import test_json_TransferToken, test_signature_TransferToken
from tests.misc.helper import get_alice_xmss, get_bob_xmss, get_slave_xmss, set_qrl_dir

logger.initialize_default()


@patch('qrl.core.txs.Transaction.logger')
class TestTransferTokenTransaction(TestCase):

    def setUp(self):
        with set_qrl_dir('no_data'):
            self.state = State()
        self.alice = get_alice_xmss()
        self.bob = get_bob_xmss()

        self.alice.set_ots_index(10)
        self.maxDiff = None

    def default_params(self):
        params = {
            "token_txhash": b'',
            "addrs_to": [self.bob.address],
            "amounts": [100],
            "fee": 1,
            "xmss_pk": self.alice.pk,
        }
        return params

    def test_create(self, m_logger):
        tx = TransferTokenTransaction.create(token_txhash=b'000000000000000',
                                             addrs_to=[self.bob.address],
                                             amounts=[200000],
                                             fee=1,
                                             xmss_pk=self.alice.pk)
        self.assertTrue(tx)

    def test_to_json(self, m_logger):
        tx = TransferTokenTransaction.create(token_txhash=b'000000000000000',
                                             addrs_to=[self.bob.address],
                                             amounts=[200000],
                                             fee=1,
                                             xmss_pk=self.alice.pk)
        txjson = tx.to_json()

        self.assertEqual(json.loads(test_json_TransferToken), json.loads(txjson))

    def test_from_json(self, m_logger):
        tx = Transaction.from_json(test_json_TransferToken)
        tx.sign(self.alice)

        self.assertIsInstance(tx, TransferTokenTransaction)

        # Test that common Transaction components were copied over.
        self.assertEqual('010300a1da274e68c88b0ccf448e0b1916fa789b01eb2ed4e9ad565ce264c9390782a9c61ac02f',
                         bin2hstr(tx.addr_from))
        self.assertEqual('01030038ea6375069f8272cc1a6601b3c76c21519455603d370036b97c779ada356'
                         '5854e3983bd564298c49ae2e7fa6e28d4b954d8cd59398f1225b08d6144854aee0e',
                         bin2hstr(tx.PK))
        self.assertEqual(b'000000000000000', tx.token_txhash)
        self.assertEqual(200000, tx.total_amount)
        self.assertEqual('390b159b34cffd29d4271a19679ff227df2ccd471078f177a7b58ca5f5d999f0', bin2hstr(tx.txhash))
        self.assertEqual(10, tx.ots_key)

        # z = bin2hstr(tx.signature)
        # print('"', end='')
        # for i in range(len(z)):
        #     print(z[i], end='')
        #     if (i + 1) % 64 == 0:
        #         print('" \\', end='')
        #         print('')
        #         print(' ' * len('test_signature_TransferToken = '), end='')
        #         print('"', end='')

        self.assertEqual(test_signature_TransferToken, bin2hstr(tx.signature))

        self.assertEqual(1, tx.fee)

    def test_validate_tx(self, m_logger):
        tx = TransferTokenTransaction.create(token_txhash=b'000000000000000',
                                             addrs_to=[self.bob.address],
                                             amounts=[200000],
                                             fee=1,
                                             xmss_pk=self.alice.pk)

        # We must sign the tx before validation will work.
        tx.sign(self.alice)

        # We have not touched the tx: validation should pass.
        self.assertTrue(tx.validate_or_raise())

    def test_validate_tx2(self, m_logger):
        params = self.default_params()
        tx = TransferTokenTransaction.create(**params)
        tx.sign(self.alice)

        self.assertTrue(tx.validate_or_raise())

        tx._data.transaction_hash = b'abc'

        # Should fail, as we have modified with invalid transaction_hash
        with self.assertRaises(ValueError):
            tx.validate_or_raise()

    def test_state_validate_tx_custom(self, m_logger):
        """
        TransferTokenTransaction._validate_custom() checks for:
        1. 0 amounts
        2. fee < 0
        3. multi-send: too many recipients
        4. multi-send: recipient addresses and amounts not the same length
        5. invalid addr_from
        6. invalid addr_to
        """
        slave = get_slave_xmss()
        params = self.default_params()
        tx = TransferTokenTransaction.create(**params)
        tx.sign(self.alice)

        result = tx.validate_or_raise()
        self.assertTrue(result)

        params = self.default_params()
        params["addrs_to"] = [self.bob.address, slave.address]
        params["amounts"] = [1, 0]
        with self.assertRaises(ValueError):
            tx = TransferTokenTransaction.create(**params)

        # Protobuf validation doesn't allow negative fees already
        params = self.default_params()
        tx = TransferTokenTransaction.create(**params)
        tx.sign(self.alice)
        with patch('qrl.core.txs.Transaction.Transaction.fee', new_callable=PropertyMock) as m_fee:
            m_fee.return_value = -1
            with self.assertRaises(ValueError):
                tx.validate_or_raise()

        # TX signing already fails if addrs_to and amounts are unequal length
        params = self.default_params()
        tx = TransferTokenTransaction.create(**params)
        tx.sign(self.alice)
        with patch('qrl.core.txs.TransferTokenTransaction.TransferTokenTransaction.addrs_to',
                   new_callable=PropertyMock) as m_addrs_to:
            m_addrs_to.return_value = [self.bob.address, slave.address]
            with self.assertRaises(ValueError):
                tx.validate_or_raise()

        params = self.default_params()
        params["master_addr"] = b'Bad QRL Address'
        with self.assertRaises(ValueError):
            tx = TransferTokenTransaction.create(**params)

        params = self.default_params()
        params["addrs_to"] = [self.bob.address, b'Bad QRL address']
        params["amounts"] = [100, 200]
        with self.assertRaises(ValueError):
            tx = TransferTokenTransaction.create(**params)

    @patch('qrl.core.txs.Transaction.Transaction.validate_slave', return_value=True)
    def test_validate_extended(self, m_validate_slave, m_logger):
        """
        TransferTokenTransaction._validate_extended checks for:
        1. valid master/slave
        2. negative fee, negative total token amounts transferred
        3. addr_from has enough funds for the fee
        4. if addr_from owns any tokens to begin with
        5. if addr_from has enough tokens
        6. addr_from ots_key reuse
        """
        alice_address_state = OptimizedAddressState.get_default(self.alice.address)
        alice_address_state.pbdata.balance = 100

        params = self.default_params()
        tx = TransferTokenTransaction.create(**params)
        tx.sign(self.alice)

        addresses_state = {
            alice_address_state.address: alice_address_state
        }
        tokens = Indexer(b'token', None)
        tokens.data[(self.alice.address, tx.token_txhash)] = qrl_pb2.TokenBalance(balance=1000,
                                                                                  decimals=0,
                                                                                  delete=False)

        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=tokens,
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        result = tx._validate_extended(state_container)
        self.assertTrue(result)

        # Invalid master XMSS/slave XMSS relationship
        m_validate_slave.return_value = False
        state_container.tokens = Indexer(b'token', None)
        state_container.tokens.data[(self.alice.address, tx.token_txhash)] = qrl_pb2.TokenBalance(balance=1000,
                                                                                                  decimals=0,
                                                                                                  delete=False)
        result = tx.validate_all(state_container)
        self.assertFalse(result)
        m_validate_slave.return_value = True

        # fee = -1
        with patch('qrl.core.txs.TransferTokenTransaction.TransferTokenTransaction.fee',
                   new_callable=PropertyMock) as m_fee:
            m_fee.return_value = -1
            tokens = Indexer(b'token', None)
            tokens.data[(self.alice.address, tx.token_txhash)] = qrl_pb2.TokenBalance(balance=1000,
                                                                                      decimals=0,
                                                                                      delete=False)
            state_container.tokens = tokens
            result = tx._validate_extended(state_container)
            self.assertFalse(result)

        # total_amount = -1
        with patch('qrl.core.txs.TransferTokenTransaction.TransferTokenTransaction.total_amount',
                   new_callable=PropertyMock) as m_total_amount:
            m_total_amount.return_value = -100
            tokens = Indexer(b'token', None)
            tokens.data[(self.alice.address, tx.token_txhash)] = qrl_pb2.TokenBalance(balance=1000,
                                                                                      decimals=0,
                                                                                      delete=False)
            state_container.tokens = tokens
            result = tx._validate_extended(state_container)
            self.assertFalse(result)

        # balance = 0, cannot pay the Transaction fee
        alice_address_state.pbdata.balance = 0
        tokens = Indexer(b'token', None)
        tokens.data[(self.alice.address, tx.token_txhash)] = qrl_pb2.TokenBalance(balance=1000,
                                                                                  decimals=0,
                                                                                  delete=False)
        state_container.tokens = tokens
        result = tx._validate_extended(state_container)
        self.assertFalse(result)
        alice_address_state.pbdata.balance = 100

        # addr_from doesn't have these tokens
        state_container.tokens = Indexer(b'token', None)
        result = tx._validate_extended(state_container)
        self.assertFalse(result)

        # addr_from doesn't have enough tokens
        tokens = Indexer(b'token', None)
        tokens.data[(self.alice.address, tx.token_txhash)] = qrl_pb2.TokenBalance(balance=99,
                                                                                  decimals=0,
                                                                                  delete=False)
        state_container.tokens = tokens
        result = tx._validate_extended(state_container)
        self.assertFalse(result)

        # addr_from_pk has used this OTS key before
        addresses_state = {
            self.alice.address: alice_address_state
        }
        state_container.addresses_state = addresses_state
        # addr_from_pk has used this OTS key before
        state_container.paginated_bitfield.set_ots_key(addresses_state, alice_address_state.address, tx.ots_key)
        result = tx.validate_all(state_container)
        self.assertFalse(result)

        slave = get_slave_xmss()
        params = self.default_params()
        params["addrs_to"] = [self.bob.address, slave.address, self.alice.address]
        params["amounts"] = [2, 3, 5]
        tx = TransferTokenTransaction.create(**params)
        tx.sign(self.alice)
        with patch('qrl.core.config', autospec=True) as m_config:
            m_config.dev = config.dev.create(config.dev.prev_state_key, config.dev.current_state_key,
                                             b'', 10, True, True)
            m_config.dev.pbdata.transaction.multi_output_limit = 1
            state_container.current_dev_config = m_config.dev
            self.assertFalse(tx._validate_extended(state_container=state_container))

    def test_set_affected_address(self, m_logger):
        result = set()
        params = self.default_params()
        tx = TransferTokenTransaction.create(**params)
        tx.set_affected_address(result)
        self.assertEqual(2, len(result))

        params = self.default_params()
        params["addrs_to"] = [self.bob.address, get_slave_xmss().address]
        params["amounts"] = [100, 200]
        tx = TransferTokenTransaction.create(**params)
        tx.set_affected_address(result)
        self.assertEqual(3, len(result))



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from mock import MagicMock
from unittest import TestCase
from pyqrllib.pyqrllib import sha2_256

from tests.misc.helper import get_slave_xmss
from qrl.core.misc import logger
from qrl.core.State import State
from qrl.core.txs.TransferTokenTransaction import TransferTokenTransaction
from qrl.core.TokenMetadata import TokenMetadata

from tests.misc.helper import set_qrl_dir, get_alice_xmss, get_bob_xmss, get_token_transaction

logger.initialize_default()

alice = get_alice_xmss()
slave = get_slave_xmss()


class TestTokenMetadata(TestCase):
    def setUp(self):
        with set_qrl_dir('no_data'):
            self.state = State()

    def test_create_token_metadata(self):
        alice_xmss = get_alice_xmss()
        bob_xmss = get_bob_xmss()

        token_transaction = get_token_transaction(alice_xmss, bob_xmss)
        TokenMetadata.create_token_metadata(self.state, token_transaction, None)

        token_metadata = TokenMetadata.get_token_metadata(self.state, token_transaction.txhash)
        self.assertEqual(token_metadata.token_txhash, token_transaction.txhash)
        self.assertEqual(token_metadata.transfer_token_tx_hashes[0], token_transaction.txhash)

    def test_update_token_metadata(self):
        alice_xmss = get_alice_xmss()
        bob_xmss = get_bob_xmss()

        token_transaction = get_token_transaction(alice_xmss, bob_xmss)
        TokenMetadata.create_token_metadata(self.state, token_transaction, None)

        transfer_token_transaction = TransferTokenTransaction.create(token_txhash=token_transaction.txhash,
                                                                     addrs_to=[alice_xmss.address],
                                                                     amounts=[100000000],
                                                                     fee=1,
                                                                     xmss_pk=bob_xmss.pk)

        TokenMetadata.update_token_metadata(self.state, transfer_token_transaction, None)

        token_metadata = TokenMetadata.get_token_metadata(self.state, token_transaction.txhash)
        self.assertEqual(len(token_metadata.transfer_token_tx_hashes), 2)
        self.assertEqual(token_metadata.transfer_token_tx_hashes[0], token_transaction.txhash)
        self.assertEqual(token_metadata.transfer_token_tx_hashes[1], transfer_token_transaction.txhash)

    def test_get_token_metadata(self):
        token_txhash = bytes(sha2_256(b'alpha'))
        token_metadata = TokenMetadata.create(token_txhash,
                                              [bytes(sha2_256(b'delta')),
                                               bytes(sha2_256(b'gamma'))])
        self.state._db.get_raw = MagicMock(return_value=token_metadata.serialize())
        self.assertEqual(TokenMetadata.get_token_metadata(self.state, token_txhash).to_json(),
                         token_metadata.to_json())

    def test_remove_transfer_token_metadata(self):
        alice_xmss = get_alice_xmss()
        bob_xmss = get_bob_xmss()

        token_transaction = get_token_transaction(alice_xmss, bob_xmss)
        TokenMetadata.create_token_metadata(self.state, token_transaction, None)

        transfer_token = TransferTokenTransaction.create(token_txhash=token_transaction.txhash,
                                                         addrs_to=[alice_xmss.address],
                                                         amounts=[100000000],
                                                         fee=1,
                                                         xmss_pk=bob_xmss.pk)
        transfer_token.sign(alice_xmss)

        TokenMetadata.update_token_metadata(self.state, transfer_token, None)
        token_metadata = TokenMetadata.get_token_metadata(self.state, transfer_token.token_txhash)
        self.assertIn(transfer_token.txhash,
                      token_metadata.transfer_token_tx_hashes)

        TokenMetadata.remove_transfer_token_metadata(self.state, transfer_token, None)
        token_metadata = TokenMetadata.get_token_metadata(self.state, transfer_token.token_txhash)
        self.assertNotIn(transfer_token.txhash,
                         token_metadata.transfer_token_tx_hashes)

    def test_remove_token_metadata(self):
        alice_xmss = get_alice_xmss()
        bob_xmss = get_bob_xmss()

        token_tx = get_token_transaction(alice_xmss, bob_xmss)
        TokenMetadata.create_token_metadata(self.state, token_tx, None)

        token_metadata = TokenMetadata.get_token_metadata(self.state, token_tx.txhash)
        self.assertEqual(token_metadata.token_txhash, token_tx.txhash)
        TokenMetadata.remove_token_metadata(self.state, token_tx, None)
        self.assertIsNone(TokenMetadata.get_token_metadata(self.state, token_tx.txhash))



from unittest import TestCase

from mock import Mock

from qrl.core import config
from qrl.core.Indexer import Indexer
from qrl.core.misc import logger
from qrl.core.State import State
from qrl.core.StateContainer import StateContainer
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.ChainManager import ChainManager
from qrl.core.txs.TransferTokenTransaction import TransferTokenTransaction
from qrl.generated.qrl_pb2 import TokenBalance
from tests.misc.helper import get_alice_xmss, get_bob_xmss, get_slave_xmss, set_qrl_dir

logger.initialize_default()


class TestTransferTokenTransactionStateChanges(TestCase):
    def setUp(self):
        with set_qrl_dir('no_data'):
            self.state = State()

        self.alice = get_alice_xmss()
        self.bob = get_bob_xmss()

        alice_address_state = OptimizedAddressState.get_default(self.alice.address)
        alice_address_state.pbdata.balance = 100
        self.addresses_state = {
            self.alice.address: alice_address_state,
            self.bob.address: OptimizedAddressState.get_default(self.bob.address)
        }

        self.params = {
            "token_txhash": b'I declare the TEST token',
            "addrs_to": [self.bob.address],
            "amounts": [100],
            "fee": 1,
            "xmss_pk": self.alice.pk
        }

        self.unused_chain_manager_mock = Mock(autospec=ChainManager, name='unused ChainManager')

    def test_apply_transfer_token_txn(self):
        """
        Alice has 1000 tokens and 100 QRL, Bob has none. Alice sends some tokens to Bob.
        """
        tx = TransferTokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = dict(self.addresses_state)
        tokens = Indexer(b'token', None)
        tokens.data[(self.alice.address, tx.token_txhash)] = TokenBalance(balance=150)
        tokens.data[(self.bob.address, tx.token_txhash)] = TokenBalance(balance=0)
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=tokens,
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))
        tx.apply(self.state, state_container)
        self.assertTrue(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))

        self.assertEqual(addresses_state[self.alice.address].balance, 99)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)

        self.assertEqual(tokens.data[(self.alice.address, tx.token_txhash)].balance, 50)
        self.assertEqual(tokens.data[(self.bob.address, tx.token_txhash)].balance, 100)

    def test_apply_transfer_token_txn_multi_send(self):
        """
        Alice has 1000 tokens and 100 QRL, Bob and Slave have none. Alice sends some tokens to Bob and Slave.
        """
        slave = get_slave_xmss()
        params = self.params.copy()
        params["addrs_to"] = [self.bob.address, slave.address]
        params["amounts"] = [100, 100]

        tx = TransferTokenTransaction.create(**params)
        tx.sign(self.alice)
        addresses_state = dict(self.addresses_state)
        addresses_state[slave.address] = OptimizedAddressState.get_default(slave.address)
        tokens = Indexer(b'token', None)
        tokens.data[(self.alice.address, tx.token_txhash)] = TokenBalance(balance=200)
        tokens.data[(self.bob.address, tx.token_txhash)] = TokenBalance(balance=0)
        tokens.data[(slave.address, tx.token_txhash)] = TokenBalance(balance=0)
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=tokens,
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)

        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))
        tx.apply(self.state, state_container)
        self.assertTrue(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))

        self.assertEqual(addresses_state[self.alice.address].balance, 99)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)
        self.assertEqual(addresses_state[slave.address].balance, 0)

        self.assertEqual(tokens.data[(self.alice.address, tx.token_txhash)].balance, 0)
        self.assertEqual(tokens.data[(self.bob.address, tx.token_txhash)].balance, 100)
        self.assertEqual(tokens.data[(slave.address, tx.token_txhash)].balance, 100)

    def test_apply_transfer_token_txn_send_tokens_to_self(self):
        """
        Alice has 1000 tokens and 100 QRL. She sends some tokens to herself. What happens next?
        """
        self.params["addrs_to"] = [self.alice.address]
        tx = TransferTokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = dict(self.addresses_state)
        tokens = Indexer(b'token', None)
        tokens.data[(self.alice.address, tx.token_txhash)] = TokenBalance(balance=100)
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=tokens,
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)

        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))
        tx.apply(self.state, state_container)
        self.assertTrue(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))

        self.assertEqual(addresses_state[self.alice.address].balance, 99)
        self.assertEqual(tokens.data[(self.alice.address, tx.token_txhash)].balance, 100)

    def test_revert_transfer_token_txn(self):
        """
        Alice has 1000 tokens and 100 QRL, Bob has none. Alice sends some tokens to Bob.
        Let's undo this.
        """
        tx = TransferTokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = dict(self.addresses_state)
        addresses_state[self.alice.address].pbdata.balance = 100
        addresses_state[self.bob.address].pbdata.balance = 0
        tokens = Indexer(b'token', None)
        tokens.data[(self.alice.address, tx.token_txhash)] = TokenBalance(balance=1000)
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=tokens,
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        state_container.paginated_bitfield.set_ots_key(addresses_state, self.alice.address, 0)

        self.assertTrue(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))
        tx.apply(self.state, state_container)
        tx.revert(self.state, state_container)
        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)

        self.assertEqual(tokens.data[(self.alice.address, tx.token_txhash)].balance, 1000)
        self.assertEqual(tokens.data[(self.bob.address, tx.token_txhash)].balance, 0)

    def test_revert_transfer_token_txn_multi_send(self):
        """
        Alice has 1100 tokens and 100 QRL, Bob and Slave have none. Alice sends some tokens to Bob and Slave.
        Undo this.
        """
        slave = get_slave_xmss()
        self.params["addrs_to"] = [self.bob.address, slave.address]
        self.params["amounts"] = [100, 100]

        tx = TransferTokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = dict(self.addresses_state)
        addresses_state[self.alice.address].pbdata.balance = 100
        addresses_state[self.bob.address].pbdata.balance = 0
        addresses_state[slave.address] = OptimizedAddressState.get_default(slave.address)
        tokens = Indexer(b'token', None)
        tokens.data[(self.alice.address, tx.token_txhash)] = TokenBalance(balance=1100)

        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=tokens,
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        state_container.paginated_bitfield.set_ots_key(addresses_state, self.alice.address, 0)

        self.assertTrue(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))
        tx.apply(self.state, state_container)
        tx.revert(self.state, state_container)
        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)
        self.assertEqual(addresses_state[slave.address].balance, 0)

        self.assertEqual(tokens.data[(self.alice.address, tx.token_txhash)].balance, 1100)
        self.assertEqual(tokens.data[(self.bob.address, tx.token_txhash)].balance, 0)
        self.assertEqual(tokens.data[(slave.address, tx.token_txhash)].balance, 0)

    def test_revert_transfer_token_txn_empty_addresses_state(self):
        """
        If we didn't have any AddressStates for the addresses involved in this test, do nothing
        """
        tx = TransferTokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = dict(self.addresses_state)
        tokens = Indexer(b'token', None)
        tokens.data[(self.alice.address, tx.token_txhash)] = TokenBalance(balance=100)
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=tokens,
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        state_container.paginated_bitfield.set_ots_key(addresses_state, self.alice.address, 0)

        self.assertTrue(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))
        tx.apply(self.state, state_container)
        tx.revert(self.state, state_container)
        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))

        self.assertEqual(100, tokens.data[(self.alice.address, tx.token_txhash)].balance)
        self.assertEqual(0, tokens.data[(self.bob.address, tx.token_txhash)].balance)

    def test_revert_transfer_token_txn_send_tokens_to_self(self):
        """
        Alice has 1000 tokens and 100 QRL. She sends some tokens to herself.
        Can we undo this?
        """
        self.params["addrs_to"] = [self.alice.address]
        tx = TransferTokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = dict(self.addresses_state)
        addresses_state[self.alice.address].pbdata.balance = 100
        tokens = Indexer(b'token', None)
        tokens.data[(self.alice.address, tx.token_txhash)] = TokenBalance(balance=100)
        state_container = StateContainer(addresses_state=addresses_state,
                                         tokens=tokens,
                                         slaves=Indexer(b'slave', None),
                                         lattice_pk=Indexer(b'lattice_pk', None),
                                         multi_sig_spend_txs=dict(),
                                         votes_stats=dict(),
                                         block_number=1,
                                         total_coin_supply=1000,
                                         current_dev_config=config.dev,
                                         write_access=True,
                                         my_db=self.state._db,
                                         batch=None)
        state_container.paginated_bitfield.set_ots_key(addresses_state, self.alice.address, 0)

        self.assertTrue(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))
        tx.apply(self.state, state_container)
        tx.revert(self.state, state_container)
        self.assertFalse(state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(self.alice.address, 0))

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        # Unfortunately importing mock.call results in some sort of ValueError so I can't check the arguments.
        self.assertEqual(tokens.data[(self.alice.address, tx.token_txhash)].balance, 100)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import argparse
import faulthandler
import os

from mock import MagicMock
from twisted.internet import reactor
from pyqrllib.pyqrllib import hstr2bin, bin2hstr

from qrl.core.AddressState import AddressState
from qrl.core.Block import Block
from qrl.core.ChainManager import ChainManager
from qrl.core.GenesisBlock import GenesisBlock
from qrl.core.misc import ntp, logger, set_logger
from qrl.core.qrlnode import QRLNode
from qrl.services.services import start_services
from qrl.core import config
from qrl.core.State import State


def parse_arguments():
    parser = argparse.ArgumentParser(description='QRL node')
    parser.add_argument('--mining_thread_count', '-m', dest='mining_thread_count', type=int, required=False,
                        default=None, help="Number of threads for mining")
    parser.add_argument('--quiet', '-q', dest='quiet', action='store_true', required=False, default=False,
                        help="Avoid writing data to the console")
    parser.add_argument('--qrldir', '-d', dest='qrl_dir', default=config.user.qrl_dir,
                        help="Use a different directory for node data/configuration")
    parser.add_argument('--no-colors', dest='no_colors', action='store_true', default=False,
                        help="Disables color output")
    parser.add_argument("-l", "--loglevel", dest="logLevel", choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                        help="Set the logging level")
    parser.add_argument('--network-type', dest='network_type', choices=['mainnet', 'testnet'],
                        default='mainnet', required=False, help="Runs QRL Testnet Node")
    parser.add_argument('--miningAddress', dest='mining_address', required=False,
                        help="QRL Wallet address on which mining reward has to be credited.")
    parser.add_argument('--mockGetMeasurement', dest='measurement', required=False, type=int, default=-1,
                        help="Warning: Only for integration test, to mock get_measurement")
    parser.add_argument('--debug', dest='debug', action='store_true', default=False,
                        help="Enables fault handler")
    parser.add_argument('--mocknet', dest='mocknet', action='store_true', default=False,
                        help="Enables default mocknet settings")
    return parser.parse_args()


def get_mining_address(mining_address: str):
    try:
        if not mining_address:
            mining_address = bytes(hstr2bin(config.user.mining_address[1:]))
        else:
            mining_address = bytes(hstr2bin(mining_address[1:]))

        if not AddressState.address_is_valid(mining_address):
            raise ValueError('Mining Address Validation Failed')

        return mining_address
    except Exception as e:
        logger.info('Failed Parsing Mining Address %s', e)

    return None


def main():
    args = parse_arguments()

    qrl_dir_post_fix = ''
    copy_files = []
    if args.network_type == 'testnet':
        config.dev.hard_fork_heights[0] = config.dev.testnet_hard_fork_heights[0]  # Hard Fork Block Height For Testnet
        qrl_dir_post_fix = '-testnet'
        package_directory = os.path.dirname(os.path.abspath(__file__))
        copy_files.append(os.path.join(package_directory, 'network/testnet/genesis.yml'))
        copy_files.append(os.path.join(package_directory, 'network/testnet/config.yml'))

    logger.debug("=====================================================================================")
    logger.info("QRL Path: %s", args.qrl_dir)
    config.user.qrl_dir = os.path.expanduser(os.path.normpath(args.qrl_dir) + qrl_dir_post_fix)
    config.create_path(config.user.qrl_dir, copy_files)
    config.user.load_yaml(config.user.config_path)

    if args.mining_thread_count is None:
        args.mining_thread_count = config.user.mining_thread_count
    logger.debug("=====================================================================================")

    config.create_path(config.user.wallet_dir)
    mining_address = None
    ntp.setDrift()

    logger.info('Initializing chain..')
    persistent_state = State()

    if args.mocknet:
        args.debug = True
        config.user.mining_enabled = True
        config.user.mining_thread_count = 1
        config.user.mining_pause = 500
        config.dev.pbdata.block.block_timing_in_seconds = 1
        config.user.genesis_difficulty = 2

        # Mocknet mining address
        # Q01050058bb3f8cb66fd90d0347478e5bdf3a475e82cfc5fe5dc276500ca21531e6edaf3d2d0f7e
        # Mocknet mining hexseed
        # 010500dd70f898c2cb4c11ce7fd85aa04554e41dcc46569871d189a3f48d84e2fbedbe176695e291e9b81e619b3625c624cde6
        args.mining_address = 'Q01050058bb3f8cb66fd90d0347478e5bdf3a475e82cfc5fe5dc276500ca21531e6edaf3d2d0f7e'

    if args.debug:
        logger.warning("FAULT HANDLER ENABLED")
        faulthandler.enable()

    if config.user.mining_enabled:
        mining_address = get_mining_address(args.mining_address)

        if not mining_address:
            logger.warning('Invalid Mining Credit Wallet Address')
            logger.warning('%s', args.mining_address)
            return False

    chain_manager = ChainManager(state=persistent_state)
    if args.measurement > -1:
        chain_manager.get_measurement = MagicMock(return_value=args.measurement)

    chain_manager.load(Block.deserialize(GenesisBlock().serialize()))

    qrlnode = QRLNode(mining_address=mining_address)
    qrlnode.set_chain_manager(chain_manager)

    set_logger.set_logger(args, qrlnode.sync_state)

    #######
    # NOTE: Keep assigned to a variable or might get collected
    admin_service, grpc_service, mining_service, debug_service = start_services(qrlnode)

    qrlnode.start_listening()

    qrlnode.start_pow(args.mining_thread_count)

    logger.info('QRL blockchain ledger %s', config.dev.version)
    if config.user.mining_enabled:
        logger.info('Mining/staking address %s using %s threads (0 = auto)', 'Q' + bin2hstr(mining_address), args.mining_thread_count)

    elif args.mining_address or args.mining_thread_count:
        logger.warning('Mining is not enabled but you sent some "mining related" param via CLI')

    reactor.run()



import logging
import threading

from qrl.core.misc import logger, logger_twisted
from qrl.core import config

LOG_FORMAT_FULL = '%(asctime)s - %(levelname)s -  %(message)s'

LOG_FORMAT_CUSTOM = '%(asctime)s|%(version)s|%(node_state)s|%(thread_id)s| ' \
                    '%(levelname)s : %(message)s'


class ContextFilter(logging.Filter):
    def __init__(self, node_state, version):
        super(ContextFilter, self).__init__()
        self.node_state = node_state
        self.version = version

    def filter(self, record):
        record.node_state = "{:<8}".format(self.node_state.state.name)
        record.version = self.version
        record.thread_id = "{:<11}".format(threading.current_thread().name)
        return True


def set_logger(args, sync_state):
    log_level = logging.INFO
    if args.logLevel:
        log_level = getattr(logging, args.logLevel)

    logger.initialize_default(force_console_output=not args.quiet).setLevel(log_level)
    custom_filter = ContextFilter(sync_state, config.dev.version)
    logger.logger.addFilter(custom_filter)
    file_handler = logger.log_to_file(config.user.log_path)
    file_handler.addFilter(custom_filter)
    file_handler.setLevel(logging.DEBUG)
    logger.set_colors(not args.no_colors, LOG_FORMAT_CUSTOM)
    logger.set_unhandled_exception_handler()
    logger_twisted.enable_twisted_log_observer()


def set_logger_default():
    log_level = logging.ERROR

    logger.initialize_default(force_console_output=False).setLevel(log_level)
    custom_filter = ContextFilter("TEST", config.dev.version)
    logger.logger.addFilter(custom_filter)
    logger.set_colors(False, LOG_FORMAT_CUSTOM)
    logger.set_unhandled_exception_handler()
    logger_twisted.enable_twisted_log_observer()
